'use strict';
/*
	The controller -- the event manager and interface between the view and model
*/

//--------------------------------------

//When doc is ready, then call "buildController"
//"buildController" must be implemented in controller subclass file
document.onreadystatechange = function () {
	if(document.readyState === 'complete') {
		try { buildController(); }
		catch(ex) {
			//ex.message
			const messages = exceptionToDisplayStrings(ex);
            //Note: unshift(a, b) -- adds set of args to beginning of array
			if (messages.length > 0 && !messages[0].includes(ex.message))
				messages.unshift(ex.message);
			const exceptionMessage = messages.join(newline());
			const consoleNote = 'See console for more information';
			if (ex.name === 'CustomException')
				alert('*Exception* \n' + ex.message + '\n' + consoleNote);
			else {
				if (!supressGeneralStartupErrors())
					alert('*Exception*\n' + exceptionMessage + '\n' + consoleNote);
			}
			throw(ex);
		}
	}
}

//--------------------------------------

//The one controller for the one view
let controller = null;

class AbstractController {

	//ivars
	//model
	//refreshableFields

	//Some browsers may not support static field so we'll declare controller above
	//static controller = null;

	constructor() {
		this.model = this.newModel();
		this.postConstruct();
		this.refresh();
		this.setInitFocus();
	}

	//----------------------------
	//Virtual - required

	newModel() {
		//Virtual Required
		implementedBySubclass(this, 'newModel');
		}

	newController() {
		//TODO -- may be obsolete
		//Virtual Required
		implementedBySubclass(this, 'newController');
	}

	//----------------------------
	//Virtual - optional

	getRefreshableFields() {
		//Virtual optional
		//e.g. subclasses should do something like this:
		//return ['count']
		return [];
	}

	attachEvents() {
		//Virtual optional
	}

	initialize() {
		//Virtual optional
	}

	//---------------------------------------------------
	//Framework, non-virtual

	postConstruct() {
		this.initialize();
		//this.attachCommonEvents();
		this.attachEvents();
		this.attachChangeEvents();
		this.attachGettingFocusEvents();

		/* needs work
		this.addElem('userTips', 'getUserTips', 'userTips', 'userTips', true);
		this.addElem('title', 'getTitle', 'title', 'title', true);
		this.addElem('userMessage', 'getUserMessage', 'userMessage', 'userMessage', false);
		*/

		this.addLabels();
		this.refreshableFields = this.getAllRefreshableFields();
		this.setWindowTitle();
	}

	getAllRefreshableFields() {
		//custom + system
		const fields = this.getRefreshableFields();
		['userTips', 'title', 'userMessage'].forEach(fieldName => {
			if (this.elem(fieldName))
				fields.push(fieldName);
		});
		return fields;
	}

	DELETEMEsetFieldSafely(fieldName, modelMessageName) {
		const elem = this.elem(fieldName);
		if (!elem) return;
		const m = this.model;
		const message = m[modelMessageName];
		if (message)
			this.setFieldValue(elem, message());
	}

	setWindowTitle() {
		const m = this.model;
		const getter = 'getTitle';
		if (m[getter])
			document.title = m[getter].bind(m)();
	}

	addLabels() {
		//class = labeled
		//TODO -- hard coded 'fields'
		const root = this.elem('fields');
		if (!root) return;
		const fieldElements = Array.from(root.getElementsByClassName(['labeled']));
		for (let field of fieldElements)
			this.addLabelTo(field);
	}

	addLabelTo(field) {
		//<label for="tries">Tries:</label>
		let label, labelElem, divElem;
		//TODO (assumption)
		//Assumption: id is good enough for label
		label = field.id.capitalized();
		divElem = document.createElement("div");
		labelElem = document.createElement("label");
		labelElem.for = label;
		labelElem.innerText = label + ':';
		divElem.append(labelElem);
		field.parentElement.parentElement.prepend(divElem);
	}

	addElem(fieldName, modelValueGetter, id, newClass, prepend) {
		//const elem = this.elem(fieldName);
		//if (!elem) return;
		/*
			<div class="row">
				<div>
					<span id=title class="title">Automatic</span>
				</div>
			</div>
		*/
		let m, message, value, fieldsElem, span, innerDiv, outerDiv;
		m = this.model;
		message = m[modelValueGetter];
		if (!message) return;
		value = message.bind(m)();
		fieldsElem = this.elem('fields');
		if (!fieldsElem) return;
		span = document.createElement("span");
		span.id = id;
		span.classList.add(newClass);
		span.innerText = value;
		innerDiv = document.createElement("div");
		outerDiv = document.createElement("div");
		outerDiv.classList.add('row');
		innerDiv.append(span);
		outerDiv.append(innerDiv);
		if (prepend)
			fieldsElem.prepend(outerDiv);
		else
			fieldsElem.append(outerDiv);
	}

	// attachCommonEvents() {
	// for (let uiElemNm of this.refreshableFields)
	// this.attach(uiElemNm, 'needsContents', `nc${uiElemNm.capitalized()}`);
	// }

	//---------------------------------------------------
	//Helper Methods

	basicAttachEventToHandler(uiElem, eventName, handlerName) {
		/*The concept here is simple enough:
			attach event for field to event handler	method
			The handler method needs context of "this" (object).
			That is what "bind" does below -- it copies the
		handler method and gives it context "this"*/
		//Attach the handler wrapper to the event
		uiElem.addEventListener(eventName, this.handlerForName(handlerName), false);
	}

	basicAttach(fieldName, eventName, handlerName) {
		if (!this.elem(fieldName))
			throw new CustomException(`Did not find field "${fieldName}" when attaching to handler "${handlerName}"`);
		this.basicAttachEventToHandler(this.elem(fieldName), eventName, handlerName);
	}

	attach(fieldName, eventName, handlerName, aRefreshAfter) {
		//NOTE: Event handlers are called in the order in which they are
		//registered, which we utilize here, e.g., {change, refresh}
		const refreshAfter = (aRefreshAfter !== undefined) ? aRefreshAfter : true;
		this.basicAttach(fieldName, eventName, handlerName);
		if (refreshAfter) {
			this.basicAttach(fieldName, eventName, 'refresh');
		}
	}

	connectEvent(fieldName, eventName, handlerName, aRefreshAfter) {
		this.attachAction(fieldName, eventName, handlerName, aRefreshAfter);
	}

	attachAction(fieldName, eventName, handlerName, aRefreshAfter) {
		const elem = this.elem(fieldName);
		if (!elem)
			throw new CustomException(`Did not find field "${fieldName}" when attaching action to handler "${handlerName}"`);
		if (!this[handlerName])
			throw new CustomException(`Did not find action handler "${handlerName}" when attaching action from element "${fieldName}"`);
		const refreshAfter = (aRefreshAfter !== undefined) ? aRefreshAfter : true;
		const actionHandler = this.handleAction.bind(this, handlerName, fieldName, eventName, refreshAfter);
		elem.addEventListener(eventName, actionHandler, false);
	}

	handleAction(handlerName, fieldName, eventName, refreshAfter) {
		try {
			this[handlerName].bind(this)(this.elem(fieldName));
		} catch(ex) {
			alert(`*Exception* during action "${eventName}" for field "${fieldName}" -- ${ex.message}`);
		}
		if (refreshAfter)
			this.refresh();
	}

	basicHandleEnterKeyEvent(handlerName, evt) {
		// "Enter" key is 13 on keyboard
		if (evt.keyCode === 13) {
			evt.preventDefault();
			this[handlerName](evt);
		}
	}

	attachEnterKeyEvent(uiElemName, handlerName) {
		this.attachEnterKeyEventTo(this.elem(uiElemName), handlerName);
	}

	attachEnterKeyEventTo(elem, handlerName) {
		const handler = this.handlerForNameWith('basicHandleEnterKeyEvent', handlerName);
		elem.addEventListener("keyup", handler, false);
	}

	OLDattachEnterKeyEvent(uiElemName, handlerName) {
		const handler = this.handlerForName(handlerName);
		this.elem(uiElemName).addEventListener("keyup", function(event) {
			// "Enter" key is 13 on keyboard
			if (event.keyCode === 13) {
				event.preventDefault();
				handler.call(this);
			}
		});
	}

	attachEnterKeyTo(handlerName) {
		const handler = this['enterKeyPressed'].bind(this, handlerName);
		document.addEventListener('keyup', handler, false);
	}

	enterKeyPressed(handlerName, evt) {
		if (event.keyCode !== 13) return;
		if (document.activeElement.nodeName.toLowerCase() !== 'body')
			return;
		event.preventDefault();
		this[handlerName].bind(this)();
	}

	refresh() {
		this.refreshFields(this.refreshableFields, true);
	}

	refreshFields(aFields, aShouldRefreshState) {
		const shouldRefreshState = aShouldRefreshState || false;
		//spec is a fieldNameSpec a pair [fieldName, getter] or just fieldName
		let fields = aFields;
		if (!Array.isArray(fields))
			fields = [aFields];
		for (let spec of fields) {
			const m = this.model;
			const fieldName = Array.isArray(spec) ? spec[0] : spec;
			//Try both 'width' or 'getWidth'

			//value = m[fieldName] !== undefined ? m[fieldName] : m['get' + fieldName.capitalized()]();
			let value = this.getModelValue(fieldName);

			//this.elem(fieldName).innerText = value;
			const elem = this.elem(fieldName);
			if (!elem)
				throw new CustomException(`Could not find html element named "${fieldName}"`);

			//TODO -- temp assumption two dec places max

			const field = this.elem(fieldName);
			if (field.readOnly)
				if ((typeof value === 'number') && (decimalCount(value) > 2))
					value = value.toFixed(2);

			this.setFieldValue(field, value);

		}
		if (shouldRefreshState)
			this.refreshState();
	}

	refreshState() {
		//virtual optional
	}

	getModelValue(fieldName) {
		const m = this.model;
		const selector = m[fieldName] !== undefined ? fieldName : 'get' + fieldName.capitalized();
		const property = m[selector];
		if ((typeof property).toLowerCase() === 'function')
			return property.bind(m)();
		return property;
	}

	setFieldValue(elem, newValue) {
		//null would seem to be a blank
		//if (newValue === null) return;
		const elemType = elem.nodeName.toLowerCase();
		if (elemType == 'input')
			elem.value = newValue;
		else
			elem.innerText = newValue;
	}

	elem(nm) {
		//Return field named 'nm'
		return document.getElementById(nm);
	}

	getElem(nm) {
		//Alias -- return field named 'nm'
		return this.elem(nm);
	}

	implementedBySubclass(nm) {
		//Throw error because subclass should have implemented method
		throw(`Method ${nm} needs to be implemented`);
	}

	handlerForName(handlerName) {
		//Return handler, we have to return a wrapper so handler has context to "this"
		if (!this[handlerName])
			throw new CustomException(`Could not find event handler method "${handlerName}"`);
		return this[handlerName + 'Wrapper'] = this[handlerName].bind(this);
	}

	handlerForNameWith(handlerName, arg) {
		//Return handler, we have to return a wrapper so handler has context to "this"
		return this[handlerName + 'Wrapper'] = this[handlerName].bind(this, arg);
	}

	//-------------------------------------------

	getChangeableFields() {
		//virtual optional
		return [];
	}

	attachChangeEvents() {
		if (!this.isAutoApply()) return;
		this.getChangeableFields().forEach(fieldName => {
			this.attach(fieldName, 'change', 'applyChanges', this.isAutoUpdate());
		});
	}

	attachGettingFocusEvents() {
		this.getChangeableFields().forEach(fieldName => {
			const field = this.elem(fieldName);
			const fieldType = field.nodeName.toLowerCase();
			if (fieldType === 'input' || fieldType === 'textarea') {
				const handler = this['inputGettingFocus'].bind(this, fieldName);
				field.addEventListener('focus', handler, false);
			}
		});
	}

	inputGettingFocus(fieldName, evt) {
		this.elem(fieldName).select();
	}

	applyChanges() {
		//virtual optional
		this.model.clearUserMessage();
		this.getChangeableFields().forEach(fieldName => {
			this.applyFieldChange(fieldName);
		});
	}

	applyFieldChange(fieldName) {
		//e.g. 'width'
		const setter = 'set' + fieldName.capitalized();
		const field = this.elem(fieldName);

		//don't need bind here
		//TODO add "perform" type protocol
		//this.model[setter].bind(this.model)(field.value);

		const m = this.model;
		const val = field.value;
		try {
			let coercedValue = null;
			if (m.validateValue(fieldName, val))
				coercedValue = m.coerce(fieldName, val);
			else
				coercedValue = m.coerce(fieldName, null);
			m[setter](coercedValue);
		} catch(ex) {
			alert(`Exception sending "${setter}('${val}')" to "${m.constructor.name}" -- ${ex.message}`);
		}
	}

	setInitFocus() {
		//default is simply first editable, if found
		const inputFields = Array.from(document.getElementsByTagName("input"));
		const fields = inputFields.concat(Array.from(document.getElementsByTagName("textarea")));
		for (let field of fields)
			if (!field.readOnly) {
				field.focus();
				field.select();
				return;
			}
	}

	isAutoUpdate() {
		//virtual optional default true
		return true;
	}

	isAutoApply() {
		//do we automatically apply changes
		return false;
	}

}

//=====================================================

function supressGeneralStartupErrors() {
	return false;
}












