
/*
algo-lab-model.js
class: AlgoLabModel

We are the model for an algorithms lab.
We have:
    zclass - the current subject class in the lab
    zmethods - (derived) the methods associated with zclass

Note:
    The 'z' prefix in zclass and zmethods is to avoid
    possible name conflic (e.g. "class" and "methods" are common)
*/

class AlgoLabModel {

    //allows us to turn the algo lab log off
    static logSwitch = true;
    static getLogSwitch() { return this.logSwitch; }
    static setLogSwitch(aSwitch) { this.logSwitch = aSwitch; }
    static getLogSwitchFct() { return onOff => this.setLogSwitch(onOff); }

    constructor(classes, initialChoice) {
        this._zclassChoices = classes;
        this.zclass(initialChoice);
        this.methodSet(new AlgoMethodSet());
        this.labSource(new AlgoLabSource());
    }

    //-------------------------------------------------------------------
    // Accessors

    methodSet(arg) {
        if (arg !== undefined) this._methodSet = arg;
        return this._methodSet;
    }

    getSource() {
        return this.labSource().source();
    }

    //-------------------------------------------------------------------
    //zclass

    coerceZclass(arg) {
        if (typeof arg === 'string')
            return this.zclassChoicesMap().get(arg);
        return arg;
    }

    zclass(aClass) {
        if (aClass !== undefined)
            this.setZclass(aClass);
        return this._zclass;
    }

    setZclass(aClassOrNm) {
        const cl = this.coerceZclass(aClassOrNm)
        this._zclass = cl;
        this.methodSet().setZclass(cl);
        this.zclassChanged();
    }

    zclassChanged() {
        this.clearDerived();
    }

    zclassChoices(choices) {
        if (choices !== undefined) this._zclassChoices = choices;
        return this._zclassChoices;
    }

    labSource(src) {
        if (src !== undefined) this._labSource = src;
        return this._labSource;
    }

    zclassChoicesMap() {
        const map = new Map();
        this.zclassChoices().forEach(cl => map.set(cl.name, cl));
        return map;
    }

    clearDerived() {
        this.methodSet().clearDerived();
        this.selectedMethodComment(null);
        this.zmethod(null);
    }

    //-------------------------------------------------------------------

    instanceOption(instanceOption) {
        return this.methodSet().instanceOption(instanceOption);
    }

    source(sourceCode) {
        return this.labSource().source(sourceCode);
    }

    selectionRange(selectionRange) {
        return this.labSource().selectionRange(selectionRange);
    }

    hasSelection() {
        return this.labSource().hasSelection();
    }

    selectionLength() {
        return this.labSource().selectionLength();
    }

    selectedText() {
        return this.labSource().getSelectedText();
    }

    errorRangeFor(error) {
        return this.labSource().errorRangeFor(error);
    }

    insert(text) {
        this.labSource().insert(text);
    }

    //-------------------------------------------------------------------
    //zmethod -- the selected method

    zmethod(arg) {
        if (arg !== undefined) {
            this._zmethod = this.constructor.cleanMethodName(arg);
            this.selectedMethodComment(null);
        }
        return this._zmethod;
    }

    //-------------------------------------------------------------------
    //zmethodComments - method comments for selected method

    selectedMethodComment(arg) {
        if (arg !== undefined) this._selectedMethodComment = arg;
        else if (this._methodComment == null) this._selectedMethodComment = this.selectedMethodCommentBlt();
        return this._selectedMethodComment;
    }

    selectedMethodCommentBlt() {
        //Return a string
        if (!this.zmethod()) return '';
        const method = this.methodSet().findMethodNamed(this.zmethod());
        if (method == null) return '';
        return this.constructor.methodHeaderCommentsFrom(method, this.zclass().name, !this.instanceOption());
    }

    //-------------------------------------------------------------------
    //deriving method header

    static methodHeaderCommentsFrom(aMethod, className, isStatic) {
        //Answer method comments (at top of method)
        const method = Array.isArray(aMethod) ? aMethod[1] : aMethod;
        const source = method.toString();
        let comments =
            (method.name === className)
                ? this.constructorMethodHeaderCommentsFromSource(source)
                : this.methodHeaderCommentsFromSource(source);
        const header = AlgoMethodSet.methodHeaderFrom(aMethod, className, isStatic);
        comments = this.cleanupComment(comments, header);
        return header  + '\n\n' + comments;
    }

    static cleanupComment(comment, header) {
        "the header may be repeated in the comment up-top (e.g., for prototypes)"
        let strmComment, strmHeader, nm;
        strmComment = TextStream.from(comment);
        strmHeader = TextStream.from(header);
        nm = strmHeader.upToFor(ch => ch === '(');
        strmComment.sw();
        if (strmComment.nextEquals(nm)) {
            strmComment.nextLine();
            strmComment.sw();
            return strmComment.copyRemaining();
        }
        return comment;
    }

    static methodHeaderCommentsFromSource(methodSource) {
        //Return leading comments
        //Assume comment lines come after method header
        //and begin with '//'
        //regex
        return this.methodHeaderCommentsFromLines(methodSource.trim().split('\n'));
    }

    static methodHeaderCommentsFromLines(lines) {
        let firstIndex, sourceIndex, commentLines, comments;
        if (lines[1].includes('/*'))
            return this.parseNextBlockCommentFromLines(lines.splice(1));
        firstIndex = lines.findIndex(line => line.trimLeft().startsWith('//'));
        if (firstIndex < 0 || firstIndex > 2) return '';
        lines = lines.slice(firstIndex);
        sourceIndex = lines.findIndex(line => !line.trimLeft().startsWith('//'));
        commentLines = lines.slice(0, sourceIndex);
        commentLines = commentLines.map(ea => ea.trim().slice('\\'.length + 1));
        comments = commentLines.join(newline());
        return comments;
    }

    static constructorMethodHeaderCommentsFromSource(methodSource) {
        //Return leading comments in constructor method
        let lines;
        lines = methodSource.trim().split('\n');
        let constructorIndex = lines.findIndex(line => line.trimLeft().startsWith('constructor('));
        lines = lines.slice(constructorIndex);
        return this.methodHeaderCommentsFromLines(lines);
    }

    static parseNextBlockCommentFromLines(lines) {
        if (lines.length === 0) return;
        const left = '/*', right = '*/';
        if (!lines[0].includes(left))
            return '';
        const i2 = lines.findIndex(line => line.includes(right));
        if (i2 < 0)
            return '';
        const comment = lines.splice(0, i2 + 1).join(newline());
        return this.trimBlockComment(comment);
    }

    static trimBlockComment(comment) {
        const left = '/*', right = '*/';
        let trimmed = comment.trim();
        if (trimmed.startsWith(left))
            trimmed = trimmed.slice(left.length);
        if (trimmed.endsWith(right))
            trimmed = trimmed.slice(0, -right.length);
        let lines = trimmed.split('\n');
        lines = lines.map(line => line.trimLeft());
        return lines.join('\n').trim();
    }

    static cleanMethodName(nm) {
        if (!nm) return nm;
        const tag = 'static ';
        if (nm.startsWith(tag))
            return nm.slice(tag.length);
        return nm;
    }

    //---------------------------------------------------------------------\
    //Public

    methodHeaders() {
        return this.methodSet().methodHeaders();
    }

    methodComment() {
        if (this.instanceMethodOption())
            return this.zmethodComments();
        else
            return this.zstaticMethodComments();
    }

    /*hasSelection() {
        const selection = this.selectedText();
        return isString(selection) && (selection.length > 2);
    }*/

    runSelection(aSeln) {
        const seln = aSeln ?? this.selectionRange();
        const sourceModel = new AlgoLabSource();
        if (!this.isSelection(seln) || !this.isValidSelection(seln))
            return SourceError.fromMessage('Invalid selection, please check selected text in source');
        sourceModel.source(this.getSource());
        sourceModel.setSelectionRange(seln);
        return this.basicRunSource(sourceModel.getSelectedText());
    }

    run() {
        //TODO -- selection (in HTML element is squirelly) -- running all
        return this.runSelection([0, this.getSource().length]);
        /*
        const src = this.getSource();
        const lineCount = (src.match(/\n/g || '') || []).length + 1;
        if (!this.hasSelection())
            if (lineCount > 5 && (src.split('\n').filter(line => line.trim().length > 0) > 5))
                return SourceError.fromMessage('Please select source code to run');
        let seln = this.hasSelection() ? this.selectionRange() : [0, src.length];
        if (isValidVarName(src.slice(seln[0], seln[1])))
            seln = [0, src.length];
        return this.runSelection(seln);
        */
    }

    basicRunSource(src) {
        const runner = SourceRunner.fromSource(src, AlgoLabModel.getLogSwitchFct());
        if (!runner.validate())
            return runner.getError();
        try {
            runner.run();
        } catch (ex) {
            return SourceError.fromMessage(ex.toString());
        }
        return runner.getError();
    }

    isSelection(seln) {
        return Array.isArray(seln);
    }

    isValidSelection(seln) {
        //seln should be at lest 5 chars
        return (seln[1] - seln[0]) >= 5;
    }

}

//-------------------------------------------------------------
AlgoLabModel;