
/*
source-runner.js
class: SourceRunner
*/

class SourceValidator {

    //---------------------------------------------------------------------------------\
    //Construction

    constructor(src, logSwitch) {
        this.source = src;
        this.logSwitch = logSwitch;
        // eslint-disable-next-line no-undef
        this.error = SourceError.newError();
        this.done = false;
    }

    static fromSource(src, logSwitch) {
        return new this(src, logSwitch);
    }

    //---------------------------------------------------------------------------------\
    //Accessors

    getError() {
        return this.error;
    }

    hasError() {
        return this.error.hasError();
    }

    setErrorMessage(errorMessage) {
        this.error.setMessage(errorMessage);
    }

    setLineNumber(no) {
        this.error.setLineNumber(no);
    }

    isDone() {
        return this.done;
    }

    //---------------------------------------------------------------------------------\
    //Public

    validate() {
        this.basicValidate();
        if (this.hasError())
            this.findErrorPosition();
        this.done = true;
    }

    //---------------------------------------------------------------------------------\
    //Private

    basicValidate() {
        //private
        //We run as (mock) file, as it allows catching errors (vanilla source code does not)
        /*unfortunately, the blob does not give us load errors for try/catch
        thus a little kludge and we wrap the entire script in function dummyOuterFct
        and run it -- NOTE WELL -- dummyOuterFct call must be in timeoutFct (i.e., queued)
        that ensures code is loaded*/
        //CHANGE -- we hoped to catch errors in a blob -- at least we can get line-numbers
        //of non-syntax errors (blob does NOT help with line #'s of  syntax errors). However
        //blob is in ui effectively (i.e., adding a HTML element) -- and ui thread is flakey,
        //e.g. need to do a setTimeout for some set time (not 0) and the time is just
        //a guess. Thus => we comment out "if (ex instanceof SyntaxError)" below
        const errorMessage = this.constructor.validateSource(this.source, this.logSwitch);
        if (errorMessage != null)
            this.setErrorMessage(errorMessage);
    }

    findErrorPosition() {
        let validateFct, finder;
        //fct returns true if valid, else false
        //validateFct = src => this.constructor.validateSource(src, this.logSwitch) == null;
        validateFct = src => this.constructor.validateSource(src, this.logSwitch);
        finder = new ErrorPositionFinder(this.getError(), this.source, validateFct);
        finder.go();
        if (finder.hasSuccess())
            this.error = finder.getError();
    }

    //-----------------------------------------------------------------------
    //Static

    static validateSource(src, logSwitchFct) {
        //private
        //return text (err msg), or null for no error
        //We run as (mock) file, as it allows catching errors (vanilla source code does not)
        /*unfortunately, the blob does not give us load errors for try/catch
        thus a little kludge and we wrap the entire script in function dummyOuterFct
        and run it -- NOTE WELL -- dummyOuterFct call must be in timeoutFct (i.e., queued)
        that ensures code is loaded*/
        //CHANGE -- we hoped to catch errors in a blob -- at least we can get line-numbers
        //of non-syntax errors (blob does NOT help with line #'s of  syntax errors). However
        //blob is in ui effectively (i.e., adding a HTML element) -- and ui thread is flakey,
        //e.g. need to do a setTimeout for some set time (not 0) and the time is just
        //a guess. Thus => we comment out "if (ex instanceof SyntaxError)" below
        if (src.length === 0) return null;
        try {
            logSwitchFct(false);
            //eval is giving better error messages than (new Function(src))();
            const src2 = this.conformSource(src);
            //eval(this.conformSource(src2));
            eval(src2);
        } catch (ex) {
            //if (ex instanceof SyntaxError)
            return ex.toString();
        } finally {
            logSwitchFct(true);
        }
        return null;
    }

    static conformSource(src) {
        /*Validation uses "eval" -- "return" statements outside of functions eval to freeze the browser
        (Chrome). So we wrap in function
        TODO -- determine if "return" statements are between braces (i.e., in a declared function) in which
        case they would be okay */
        if (!removeCppComments(src).includes('return'))
            return src;
        let src2 = src.trimRight();
        const ch = src2[src2.length - 1];
        if (ch !== ';' && ch != '}')
            src2 += ';';
        return `(function () {\n ${src2} \n})();`;
    }


}

//-------------------------------------------------------------------------------

/*
Pass #1 -- assume a very small number of source lines, e.g., < 10
*/

class ErrorPositionFinder {

    constructor(error, src, validateFct) {
        this.source = src;
        this.validateFct = validateFct;
        // eslint-disable-next-line no-undef
        this.error = SourceError.fromMessage(error.getMessage());
        this.commented = null;
    }

    go() {
        this.basicGo();
        this.conform();
    }

    getError() {
        return this.error;
    }

    hasSuccess() {
        return this.getError().hasLineNumber();
    }

    //-----------------------------------------------------

    /*
    TODO - failing case with terminating straggline "}"
    try to fix via:
        basicGo() {
            //...
            src = this.source.trimRight();
            originalLines = src.split('\n');
            lines = [...originalLines];
            lines = this.constructor.whiteoutCommentsFrom(lines);
            basicGoWith(originalLines, lines, 1)
            //...
        }
        basicGoWith(originalLines, lines, tries) {
            If tries <= 0, return
            If we do not find a line, e.g., "//Assume Line 1"
            then recurse (originalLines, lines, tries - 1)
        }

    */

    basicGo() {
        let src, originalLines, lines, validateFct, err, lineIndex, prevError;
        src = this.source.trimRight();
        //src = this.sourceNoComments();
        //this.commented = src.length < this.source.length;
        originalLines = src.split('\n');
        //spread operator copies array
        lines = [...originalLines];
        lines = this.constructor.whiteoutCommentsFrom(lines);
        validateFct = this.validateFct;
        err = this.error;
        lineIndex = lines.length - 1;
        //while (lines.length >= 0) {
        prevError = null;
        while (lineIndex >= 0) {
            const currentError = validateFct(lines.join('\n'));
            if (!currentError) {
                //const lineNumber = lines.length + 1;
                // add 1 to make one-based, and 1 because error is one line beyond
                const lineNumber = lineIndex + 1 + 1;
                this.lineNumber = lines.length + 1;
                err.setMessage(prevError);
                err.setLineNumber(lineNumber);
                err.setLineText(originalLines[lineNumber - 1]); //zero-based index
                return;
            }
            //lines.splice(lines.length - 1, 1);
            /*
            let line = lines[lineIndex];
            if (!line.includes('{') && !line.includes('}'))
                lines[lineIndex] = '// ' + line;
            */
            this.reduce(lines, lineIndex);
            lineIndex--;
            prevError = currentError;
        }
        //Assume Line 1
        err.setLineNumber(1);
        err.setLineText(originalLines[0]);
    }

    reduce(lines, lineIndex) {
        //comment out the last line, or in special cases multiple lines,
        //as we are working backwards up the
        //source looking for the error

        let line = lines[lineIndex];

        if (this.reduceOneLineControlBlock(lines, lineIndex))
            return;

        //special case (straggling '}')
        if (line.includes('}') && lines.slice(0, lineIndex).filter(ea => ea.includes('{')).length === 0) {
            lines[lineIndex] = '// ' + line;
            return;
        }

        //special case -- if on "{" then comment out it with following "}"
        if (line.includes('{')) {
            if (line.includes('}'))
                return;
            const remaining = lines.slice(lineIndex + 1);
            const remainingIndex = remaining.findIndex(ea => ea.includes('}'));
            if (remainingIndex >= 0) {
                const i = lineIndex + 1 + remainingIndex;
                lines[i] = '// ' + lines[i];
            }
        }
        if (!line.includes('}'))
            lines[lineIndex] = '// ' + line;
    }

    indexOfPrevNonBlankLine(lines, lineIndex) {
        if (lineIndex <= 0) return -1;
        for (let i = lineIndex - 1; i >= 0; i--)
            if (lines[i].trim().length > 0)
                return i;
        return -1
    }

    reduceOneLineControlBlock(lines, lineIndex) {
        //special case, if previous line is control (if, while, for)
        //and there is not "{" == then we have one-line code block
        //after control. Ccomment out current line and previous line
        let prevIndex, prevLine, keywords;
        if (lineIndex < 1) return false;
        prevIndex = this.indexOfPrevNonBlankLine(lines, lineIndex);
        if (prevIndex < 0) return false;
        prevLine = lines[prevIndex];
        keywords = ['if', 'while', 'for'];
        //okay we have control line, now do we have "{"?
        if (prevLine.includes('{')) return false;
        if (['if', 'while', 'for'].findIndex(eaKeyword => prevLine.includes(eaKeyword)) >= 0) {
            lines[prevIndex] = '// ' + prevLine;
            lines[lineIndex] = '// ' + lines[lineIndex];
            return true;
        }
        return false;
    }

    conform() {
        //if we have comments, we need to re-derive result
        let lines, err, errorLineText;
        if (!this.commented) return;
        lines = this.source.split('\n');
        err = this.error;
        errorLineText = err.getLineText();
        if (!errorLineText) return;
        errorLineText = errorLineText.trim();
        for (let i = 0; i < lines.length; i++) {
            const line = lines[i];
            if (line === errorLineText) {
                err.setLineNumber(i + 1);
                return;
            }
        }
        //Should not get here
        throw 'Could not re-find error';
    }

    static whiteoutCommentsFrom(src) {
        let newLines, lineCommentPattern, lineCommentRegex,
            inMultilineComment, left, right, isOpenerOnly, hasCloser;
        newLines = [];
        //TODO -- SEE IF WE CAN remove escape here
        // eslint-disable-next-line no-useless-escape
        lineCommentPattern = '\/\/.*$';
        lineCommentRegex = new RegExp(lineCommentPattern);
        inMultilineComment = false;
        left = '/*';
        right = '*/';
        isOpenerOnly = l => l.includes(left) && !l.includes(right);
        hasCloser = l => l.includes(right);
        //for block comments, for the context of looking for full line matches
        //Note: we only care about whiting out the in-between comments (on lines w/o /* or */)
        for (let line of src) {
            let newLine = line.replace(lineCommentRegex, '');
            if (inMultilineComment) {
                if (hasCloser(line))
                    inMultilineComment = false;
                else
                    newLine = '';
            }
            else
                inMultilineComment = isOpenerOnly(line);
            newLines.add(newLine);
        }
        return newLines;
    }

    sourceNoComments() {
        return removeCppComments(this.source);
    }

}



//-------------------------------------------------------------
SourceValidator;
ErrorPositionFinder;

























